import torch
import torch.nn as nn
from model.backbone import MyInception_v3
from roi_align.roi_align import RoIAlign  # RoIAlign module
import torch.nn.functional as F
from model.spatial_attention import Encoder_Blocks, temporal_position_encoding


class Formation_Detector(nn.Module):

    def __init__(self, args):
        super(Formation_Detector, self).__init__()
        # Set parameters
        self.args = args
        self.emb_dim = 1024
        D = 1056  # output feature map channel of backbone
        K = self.args.crop_size[0]  # crop size of roi align

        # Set networks
        self.backbone = MyInception_v3(transform_input=False, pretrained=True)
        self.roi_align = RoIAlign(*self.args.crop_size)
        self.linear = nn.Linear(K * K * D, self.emb_dim)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=self.args.train_dropout_prob)
        self.spatial_attn = Encoder_Blocks(self.emb_dim, args.num_heads, args.num_layers)
        self.fc_cls = nn.Linear(self.emb_dim, 1)
        self.sigmoid = nn.Sigmoid()
        self.fc_pos = nn.Linear(self.emb_dim, 2)

    def boxes_calculate(self, args):
        OH, OW = args.out_size
        stride_x = OW / args.num_anchors
        stride_y = OH / args.num_anchors
        boxes_in_list = []
        for i in range(args.num_anchors):
            for j in range(args.num_anchors):
                boxes_in_list.append(torch.tensor([j*stride_x, i*stride_y, (j+1)*stride_x, (i+1)*stride_y]).unsqueeze(0))  # x1,y1,x2,y2; 1,4
        boxes_in_flat = torch.cat(boxes_in_list, dim=0)  # N,4
        return boxes_in_flat

    def savemodel(self, path):
        torch.save(self.state_dict(), path)
        print(f'Model is saved at {path} successfully!')

    def loadmodel(self, path):
        self.load_state_dict(torch.load(path))
        print(f'Load model from {path} successfully!')

    def forward(self, img):
        # Set parameters
        args = self.args
        bs = img.shape[0]
        N = args.num_anchors ** 2
        H, W = args.size_of_img
        OH, OW = args.out_size

        # CNN backbone
        cnn_feature = self.backbone(img)
        features_multiscale = []
        for features in cnn_feature:
            if features.shape[2:4] != torch.Size([OH, OW]):
                features = F.interpolate(features, size=(OH, OW), mode='bilinear', align_corners=True)
            features_multiscale.append(features)
        features_multiscale = torch.cat(features_multiscale, dim=1)  # B, D, OH, OW

        # RoIAlign
        boxes_in_flat = torch.cat([self.boxes_calculate(args).unsqueeze(0) for i in range(bs)], dim=0).reshape(-1, 4).to(img.device)  # B*N,4
        boxes_idx_flat = torch.stack([i * torch.ones(N, dtype=torch.int) for i in range(bs)]).reshape(-1).to(img.device)  # B*N
        boxes_in_flat.requires_grad = False
        boxes_idx_flat.requires_grad = False
        boxes_features = self.roi_align(features_multiscale, boxes_in_flat, boxes_idx_flat)  # B*N, D, K, K,

        # Embedding
        boxes_features = boxes_features.reshape(bs * N, -1)  # B*N, D*K*K
        boxes_features = self.linear(boxes_features)  # B*N,1024
        boxes_features = self.relu(boxes_features)
        boxes_features = self.dropout(boxes_features)
        boxes_features = boxes_features.reshape(bs, N, self.emb_dim)  # B,N,1024

        # Spatial attention
        boxes_features = boxes_features + temporal_position_encoding(boxes_features.shape).to(img.device)
        boxes_features = self.spatial_attn(boxes_features)  # B,N,1024

        # Predict
        cls = self.sigmoid(self.fc_cls(boxes_features)).squeeze(2)  # B,N
        pos = self.fc_pos(boxes_features)  # B,N,2

        return cls, pos, boxes_features
